part of mercury;

/// A mapping of snowflake IDs to Discord entities.
///
/// Based off of the [Collection] implemenetation at [abalabahaha/eris]
/// https://github.com/abalabahaha/eris/blob/master/lib/util/Collection.js
class _EntityCollection<TEntity extends EntityBase> {
  /// The internal mapping of IDs to entities.
  Map<String, TEntity> _map = null;

  /// Returns the internal map store.
  Map<String, TEntity> getBaseMap() => _map;

  /// Returns the keys of this collection.
  Iterable<String> get keys => _map.keys;

  /// Returns the values of this collection.
  Iterable<TEntity> get values => _map.values;

  /// A limit of how many entities can be in this store.
  int limit = null;

  /// Whether this collection has a limit.
  bool get hasLimit => limit != null;

  /// Whether this collection does not have a limit.
  bool get hasNoLimit => !hasLimit;

  /// Whether this collection is holding the maximum amount of entities it is limited to.
  bool get isFull {
    if (hasNoLimit) return false;
    return _map.length == limit;
  }

  /// Whether this collection is not holding the maximum amount of entities it is limited to.
  bool get isNotFull => !isFull;

  _EntityCollection([this._map = null, this.limit = null]) {
    if (this._map == null) {
      this._map = <String, TEntity>{};
    }
  }

  /// Sets the mapping for ID [entity.id] to [entity].
  void set(TEntity entity, {bool replaceExisting = true}) {
    if (isFull) return;
    if (entity.id == null)
      throw new CollectionException(
          'Cannot insert an object with no ID into cache', this);

    var existing = get(entity.id);

    if (existing != null && !replaceExisting) return;
    _map[entity.id] = entity;
  }

  /// Gets an entity with ID [id] from this collection.
  EntityCollectionEntry<TEntity> get(String id) {
    return new EntityCollectionEntry(_map[id], this);
  }

  /// Returns whether a certain entity exists in this collection.
  bool containsEntity(TEntity entity) => containsId(entity.id);

  /// Returns whether a certain entity does not exist in this collection.
  bool doesNotContainEntity(TEntity entity) => !containsEntity(entity);

  /// Returns whether a certain ID exists in this collection.
  bool containsId(String id) => _map.containsKey(id);

  /// Returns whether a certain ID does not exist in this collection.
  bool doesNotContainId(String id) => !containsId(id);

  /// Returns the first entities to meet the specified predicate.
  ///
  /// Returns null if no entities match the predicate.
  TEntity firstWhere(bool Function(String id, TEntity entity) test) {
    _map.forEach((id, ent) {
      if (test(id, ent)) {
        return ent;
      }
    });
    return null;
  }

  /// Returns all entities that match the specified predicate.
  ///
  /// Returns null if no entities match the predicate.
  List<TEntity> allWhere(bool Function(String id, TEntity entity) test) {
    var newMap = <TEntity>[];
    _map.forEach((id, ent) {
      if (test(id, ent)) {
        newMap.add(ent);
      }
    });
    if (newMap.isEmpty) return null;
    return newMap;
  }

  /// Removes all entities that match the specified predicate.
  void removeWhere(bool Function(String id, TEntity entity) test) {
    var entities = allWhere(test);
    for (var ent in entities) {
      remove(ent.id);
    }
  }

  /// Maps all entities in this collection into a new [_EntityCollection] based on the specified predicate.
  _EntityCollection<TEntity> map(
      MapEntry<String, TEntity> Function(String id, TEntity entity) morph) {
    return new _EntityCollection(_map.map(morph), this.limit);
  }

  /// Removes the entity with ID [id].
  void remove(String id) {
    _map.remove(id);
  }
}

/// An enumeration for different cache model update types
enum EntityChange { Updated, Added, Deleted }

/// An entry from an [_EntityCollection]
class EntityCollectionEntry<TEntity extends EntityBase> {
  TEntity value;
  _EntityCollection collection;
  EntityCollectionEntry(this.value, this.collection);

  EntityCollectionEntry set(TEntity newEntity, {bool replaceExisting: true}) {
    collection.set(newEntity, replaceExisting: replaceExisting);
    return collection.get(newEntity.id);
  }

  void remove() {
    collection.remove(value.id);
  }
}
